/*
Copyright © 2025 filifa

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
package markov

import (
	"errors"
	"math"
	"strconv"

	"gonum.org/v1/gonum/graph"
	"gonum.org/v1/gonum/graph/encoding"
	"gonum.org/v1/gonum/graph/multi"
	"gonum.org/v1/gonum/graph/simple"
	"gonum.org/v1/gonum/mat"
)

type AbsorbingMarkovChain struct {
	*multi.WeightedDirectedGraph
}

func NewAbsorbingMarkovChain() *AbsorbingMarkovChain {
	return &AbsorbingMarkovChain{WeightedDirectedGraph: multi.NewWeightedDirectedGraph()}
}

func (g *AbsorbingMarkovChain) IsValid() bool {
	for nodes := g.Nodes(); nodes.Next(); {
		u := nodes.Node().(*node)
		if g.outWeightSum(u) != 1 {
			return false
		}
	}

	return true
}

func (g *AbsorbingMarkovChain) outWeightSum(u graph.Node) float64 {
	sum := 0.0
	for nodes := g.From(u.ID()); nodes.Next(); {
		v := nodes.Node()
		e := g.WeightedEdge(u.ID(), v.ID())
		if e != nil {
			sum += e.Weight()
		}
	}

	return sum
}

func (g *AbsorbingMarkovChain) AdjacencyMatrix() mat.Matrix {
	adj := simple.NewDirectedMatrix(g.Nodes().Len(), 0, 0, 0)
	for edges := g.WeightedEdges(); edges.Next(); {
		e := edges.WeightedEdge()
		if e.From() == e.To() {
			continue
		}

		adj.SetWeightedEdge(e)
	}

	a := mat.DenseCopyOf(adj.Matrix())

	nodes := adj.Nodes()
	for i := 0; nodes.Next(); i++ {
		id := nodes.Node().ID()
		u := g.Node(id).(*node)

		e := g.WeightedEdge(u.ID(), u.ID())
		if e != nil {
			a.Set(i, i, e.Weight())
		}
	}

	return a
}

func (g *AbsorbingMarkovChain) NewEdge(from, to graph.Node) graph.Edge {
	e := g.WeightedDirectedGraph.NewWeightedLine(from, to, math.NaN()).(multi.WeightedLine)
	return &weightedEdge{WeightedLine: e}
}

func (g *AbsorbingMarkovChain) NewNode() graph.Node {
	return &node{Node: g.WeightedDirectedGraph.NewNode()}
}

func (g *AbsorbingMarkovChain) SetEdge(e graph.Edge) {
	g.WeightedDirectedGraph.SetWeightedLine(e.(*weightedEdge))
}

type weightedEdge struct {
	multi.WeightedLine
}

func (e *weightedEdge) ReversedEdge() graph.Edge {
	revLine := multi.WeightedLine{F: e.T, T: e.F, W: e.W}
	return &weightedEdge{WeightedLine: revLine}
}

func (e *weightedEdge) SetAttribute(attr encoding.Attribute) error {
	var err error

	switch attr.Key {
	case "weight":
		e.W, err = strconv.ParseFloat(attr.Value, 64)
	default:
		err = errors.New("unknown key" + attr.Key)
	}

	return err
}

type node struct {
	graph.Node
	dotID string
}

func (n *node) SetDOTID(id string) {
	n.dotID = id
}

func (n *node) DOTID() string {
	return n.dotID
}
